
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. Both must be integers.
#define PPWM  800         // this gives 50Hz when we use the 200 and 800 combination.
      
uint8_t v1low,uf;
uint16_t pcount;
uint32_t vpwr;
uint16_t lookup_sine[NPWM+1];
uint8_t dbounce;
float pwr;

void setup()
  {
  int i;
  float t,u;

  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   
    u = 65535.0 * sin(t);   // 1/2 wave sine lookup table, scaled to (16 bits - 1)
    lookup_sine[i] =  (int)u;  
    }
  pinMode(8,INPUT);       // initialise 5 pins of nano... run/stop input command from nano2
  pinMode(9, OUTPUT);     // ..V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(12, OUTPUT);    // inverter ON/OFF LED, pin 15 on nano, Arduino D12
  pinMode(5,OUTPUT);      // used for IR2110 shutdown drive, LOW = shutdown
  
  cbi(PORTD,5);     // pull IR2184 shutdown LOW to disable gate drive output
  pwr = 0.0;        // zero the power command
  uf=0;             // initialise the 1/2 wave toggle flag to a known value
  
  noInterrupts();     // set up the 20kHz PWM interrupt, I write directly to the ATMega328 TIMER1 registers here.
  v1low = 0;             // start at 2nd half wave
  TCCR1A =  _BV(COM1A1) | _BV(WGM11); // config output compare to make pin 9 reflect PWM output and force pin 10 high for the half cycle  
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10); // define timer clock source and PWM type.
  OCR1A = 10;               // some value, not needed really. Just there as a marker for people who simply copy the code but don't understand it.
  ICR1 = PPWM;              // the main interrupt period value, in clocks. 800 clocks @ 16Mhz = 20kHz
  TIMSK1 |= (1 << TOIE1);
  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  }


// The following runs at 20kHz, think of it as the system clock.
// it makes the square wave, with a duty cycle follwing a sine wave, code has only 50uS to run, takes about 10uS
//
ISR(TIMER1_OVF_vect)        
 {   
  long c;
  c = (lookup_sine[pcount] * vpwr) >> 16;   // scale sine wave by vpwr, 32 bit integer calcs. 16 bit shift fastest of all
  if (v1low == 1)                 // alternate between 2 output compare pins. Get a full 50Hz waveform
    OCR1B = c;                    //  1/2 out this pin
  else
    OCR1A = c;                    // and then 1/2 out this pin. 
  pcount++;                       // this is the index, used as we work our way through the 200 entry sine wave look up table.
  
  if(pcount >= NPWM)          // pcount  will = NPWM, at each 1/2 wave, at the start
    {
    pcount = 0;               // reset counter to begining of the 1/2 sine wave lookup table location
    uf=1;                     // enable some code in loop() to execute at this time being zero crossing
    if (v1low == 1)           // if first 1/2 wave..
      {
       v1low = 0;             // toggle to 2nd half wave
       TCCR1A =  _BV(COM1A1) | _BV(WGM11); // config output compare to make pin 9 reflect PWM output and force pin 10 high for the half cycle  
       sbi(PORTD,7);     // send out pin 7 something for the DSO to trigger on. This will be high for one 1/2 wave, low for the other                      
      }                         
    else
      {
       v1low = 1;             // toggle to 1st half wave
       TCCR1A =  _BV(COM1B1) | _BV(WGM11);   // config O.C. to make pin 10 reflect PWM output and force pin 9 high for the half cycle.
       cbi(PORTD,7);     //DSO low 
      }
    }
}



void loop()
{
if (uf == 1)
  {
  uf = 0;             // set to 0 to prevent running this code again until 1/100 second passes  
  pwr = pwr * 0.95 + 0.05 * (float)analogRead(4) / 1024.0; // scale 0 - 5V to 0.0 to 1.0 and low pass it with shitty but simple filter
  if (pwr > 0.99) pwr = 0.99;     // clamp. Do math in floating point to prevent any integer overflow
  if (pwr < 0.01) pwr = 0.01;     // it's slower. We could do clamp after conversion to int but it's your mosfets...
  vpwr = (int)((float)PPWM * pwr);// apply required PWM duty power factor to vpwr (which will only be from 8 to 792 now.)    
  if ((PINB & 0x01) == 0x01)      // this comes from nano2, the command to run the inverter. Active high. Is it high?
    {
    sbi(PORTD,5);     // pull pin5 HIGH, to enable gate drive output by driving IR21844 shutdown pin high. (shutdown active low.)
    sbi(PORTB,4);     // and light LED on D12, pin 15 of nano
    }
  else
    {
    cbi(PORTD,5);     // pull it LOW, shutdown gate drive outputs
    cbi(PORTB,4);     // and dim LED 
    }
  }   
}
